@@ -1,6 +1,6 @@
 import functools
 from enum import Enum
-from typing import List, Optional, Tuple, Union
+from typing import Any, List, Optional, Sequence, TYPE_CHECKING, Tuple, Union
 
 import numpy as np
 import tensorflow as tf
@@ -10,30 +10,56 @@ from typing_extensions import Final
 
 from .config import default_float, default_summary_fmt
 
+if TYPE_CHECKING:
+    from IPython.lib import pretty
+
 DType = Union[np.dtype, tf.DType]
-VariableData = Union[List, Tuple, np.ndarray, int, float]
+VariableData = Union[List, Tuple, np.ndarray, int, float]  # deprecated
 Transform = Union[tfp.bijectors.Bijector]
 Prior = Union[tfp.distributions.Distribution]
 
 
-def _IS_PARAMETER(o):
+TensorType = Union[tf.Tensor, tf.Variable, "Parameter"]
+"""
+Type alias for tensor-like types that are supported by most TensorFlow and GPflow operations.
+
+NOTE: Union types like this do not work with the `register` method of `multipledispatch`'s
+`Dispatcher` class. Instead use `TensorLike`.
+"""
+
+
+# We've left this as object until we've tested the performance consequences of using the full set
+# (np.ndarray, tf.Tensor, tf.Variable, Parameter), see https://github.com/GPflow/GPflow/issues/1434
+TensorLike: Final[Tuple[type, ...]] = (object,)
+"""
+:var TensorLike: Collection of tensor-like types for registering implementations with
+    `multipledispatch` dispatchers.
+"""
+
+
+_NativeScalar = Union[int, float]
+_Array = Sequence[Any]  # a nested array of int, float, bool etc. kept simple for readability
+TensorData = Union[_NativeScalar, _Array, TensorType]
+
+
+def _IS_PARAMETER(o: Any) -> bool:
     return isinstance(o, Parameter)
 
 
-def _IS_TRAINABLE_PARAMETER(o):
+def _IS_TRAINABLE_PARAMETER(o: Any) -> bool:
     return _IS_PARAMETER(o) and o.trainable
 
 
 class Module(tf.Module):
     @property
-    def parameters(self):
+    def parameters(self) -> Tuple["Parameter", ...]:
         return tuple(self._flatten(predicate=_IS_PARAMETER))
 
     @property
-    def trainable_parameters(self):
+    def trainable_parameters(self) -> Tuple["Parameter", ...]:
         return tuple(self._flatten(predicate=_IS_TRAINABLE_PARAMETER))
 
-    def _representation_table(self, object_name, tablefmt):
+    def _representation_table(self, object_name: str, tablefmt: Optional[str]) -> str:
         from .utilities import leaf_components, tabulate_module_summary
 
         repr_components = [object_name]
@@ -41,13 +67,13 @@ class Module(tf.Module):
             repr_components.append(tabulate_module_summary(self, tablefmt=tablefmt))
         return "\n".join(repr_components)
 
-    def _repr_html_(self):
+    def _repr_html_(self) -> str:
         """ Nice representation of GPflow objects in IPython/Jupyter notebooks """
         from html import escape
 
         return self._representation_table(escape(repr(self)), "html")
 
-    def _repr_pretty_(self, p, cycle):
+    def _repr_pretty_(self, p: "pretty.RepresentationPrinter", cycle: bool) -> None:
         """ Nice representation of GPflow objects in the IPython shell """
         repr_str = self._representation_table(repr(self), default_summary_fmt())
         p.text(repr_str)
@@ -61,7 +87,7 @@ class PriorOn(Enum):
 class Parameter(tf.Module):
     def __init__(
         self,
-        value,
+        value: TensorData,
         *,
         transform: Optional[Transform] = None,
         prior: Optional[Prior] = None,
@@ -81,7 +107,7 @@ class Parameter(tf.Module):
 
         self._transform = transform
         self.prior = prior
-        self.prior_on = prior_on
+        self.prior_on = prior_on  # type: ignore  # see https://github.com/python/mypy/issues/3004
 
         if isinstance(value, tf.Variable):
             self._unconstrained = value
@@ -91,7 +117,7 @@ class Parameter(tf.Module):
                 unconstrained_value, dtype=dtype, name=name, trainable=trainable
             )
 
-    def log_prior_density(self):
+    def log_prior_density(self) -> tf.Tensor:
         """ Log of the prior probability density of the constrained variable. """
 
         if self.prior is None:
@@ -116,41 +142,41 @@ class Parameter(tf.Module):
             return log_p
 
     @property
-    def prior_on(self):
+    def prior_on(self) -> PriorOn:
         return self._prior_on
 
     @prior_on.setter
-    def prior_on(self, value: Union[str, PriorOn]):
+    def prior_on(self, value: Union[str, PriorOn]) -> None:
         self._prior_on = PriorOn(value)
 
-    def value(self):
-        return _to_constrained(self._unconstrained.value(), self.transform)
+    def value(self) -> tf.Tensor:
+        return _to_constrained(self._unconstrained.value(), self.transform)  # type: ignore  # assumes _to_constrained returns a tf.Tensor
 
-    def read_value(self):
-        return _to_constrained(self._unconstrained.read_value(), self.transform)
+    def read_value(self) -> tf.Tensor:
+        return _to_constrained(self._unconstrained.read_value(), self.transform)  # type: ignore  # assumes _to_constrained returns a tf.Tensor
 
-    def experimental_ref(self):
+    def experimental_ref(self) -> "Parameter":
         return self
 
-    def deref(self):
+    def deref(self) -> "Parameter":
         return self
 
     @property
-    def unconstrained_variable(self):
+    def unconstrained_variable(self) -> tf.Variable:
         return self._unconstrained
 
     @property
-    def transform(self):
+    def transform(self) -> Optional[Transform]:
         return self._transform
 
     @transform.setter
-    def transform(self, new_transform):
+    def transform(self, new_transform: Optional[Transform]) -> None:
         constrained_value = self.read_value()
         self._transform = new_transform
         self.assign(constrained_value)
 
     @property
-    def trainable(self):
+    def trainable(self) -> bool:
         """
         `True` if this instance is trainable, else `False`.
 
@@ -159,10 +185,10 @@ class Parameter(tf.Module):
         return self._unconstrained.trainable
 
     @property
-    def initial_value(self):
+    def initial_value(self) -> tf.Tensor:
         return self._unconstrained.initial_value
 
-    def validate_unconstrained_value(self, value: tf.Tensor, dtype: DType) -> tf.Tensor:
+    def validate_unconstrained_value(self, value: TensorData, dtype: DType) -> tf.Tensor:
         value = _cast_to_dtype(value, dtype)
         unconstrained_value = _to_unconstrained(value, self.transform)
         message = (
@@ -173,8 +199,12 @@ class Parameter(tf.Module):
         return tf.debugging.assert_all_finite(unconstrained_value, message=message)
 
     def assign(
-        self, value: tf.Tensor, use_locking=False, name=None, read_value=True
-    ) -> tf.Variable:
+        self,
+        value: TensorData,
+        use_locking: bool = False,
+        name: Optional[str] = None,
+        read_value: bool = True,
+    ) -> tf.Tensor:
         """
         Assigns constrained `value` to the unconstrained parameter's variable.
         It passes constrained value through parameter's transform first.
@@ -201,7 +231,7 @@ class Parameter(tf.Module):
         )
 
     @property
-    def is_tensor_like(self):
+    def is_tensor_like(self) -> bool:
         """
         This method means that TensorFlow's `tensor_util.is_tensor` function
         will return `True`
@@ -209,47 +239,47 @@ class Parameter(tf.Module):
         return True
 
     @property
-    def name(self):
+    def name(self) -> str:
         return self._unconstrained.name
 
     @property
-    def initializer(self):
+    def initializer(self):  # type unknown
         return self._unconstrained.initializer
 
     @property
-    def device(self):
+    def device(self) -> Optional[str]:
         return self._unconstrained.device
 
     @property
-    def dtype(self):
+    def dtype(self) -> tf.DType:
         return self._unconstrained.dtype
 
     @property
-    def op(self):
+    def op(self) -> tf.Operation:
         return self._unconstrained.op
 
     @property
-    def shape(self):
+    def shape(self) -> tf.TensorShape:
         if self.transform is not None:
             return self.transform.forward_event_shape(self._unconstrained.shape)
         return self._unconstrained.shape
 
-    def numpy(self):
+    def numpy(self) -> np.ndarray:
         return self.read_value().numpy()
 
-    def get_shape(self):
+    def get_shape(self) -> tf.TensorShape:
         return self.shape
 
-    def _should_act_as_resource_variable(self):
+    def _should_act_as_resource_variable(self):  # type unknown
         # needed so that Parameters are correctly identified by TensorFlow's
         # is_resource_variable() in resource_variable_ops.py
         pass  # only checked by TensorFlow using hasattr()
 
     @property
-    def handle(self):
+    def handle(self):  # type unknown
         return self._unconstrained.handle
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         unconstrained = self.unconstrained_variable
         constrained = self.read_value()
         if tf.executing_eagerly():
@@ -314,26 +344,13 @@ class Parameter(tf.Module):
 Parameter._OverloadAllOperators()
 tf.register_tensor_conversion_function(Parameter, lambda x, *args, **kwds: x.read_value())
 
-TensorType = Union[np.ndarray, tf.Tensor, tf.Variable, Parameter]
-"""
-Type alias for tensor-like types that are supported by most TensorFlow, NumPy and GPflow operations.
 
-NOTE: Union types like this do not work with the `register` method of multipledispatch's
-`Dispatcher` class. Instead use `TensorLike` for dispatching on tensor-like types.
-"""
-
-# We've left this as object until we've tested the performance consequences of using the full set
-# (np.ndarray, tf.Tensor, tf.Variable, Parameter), see https://github.com/GPflow/GPflow/issues/1434
-TensorLike: Final[Tuple[type, ...]] = (object,)
-"""
-:var TensorLike: Collection of tensor-like types for registering implementations with
-    `multipledispatch` dispatchers.
-"""
-
-
-def _cast_to_dtype(value: VariableData, dtype: Optional[DType] = None) -> tf.Tensor:
+def _cast_to_dtype(
+    value: TensorData, dtype: Optional[DType] = None
+) -> Union[tf.Tensor, tf.Variable]:
     if dtype is None:
         dtype = default_float()
+
     if tf.is_tensor(value):
         # NOTE(awav) TF2.2 resolves issue with cast.
         # From TF2.2, `tf.cast` can be used alone instead of this auxiliary function.
@@ -343,13 +360,13 @@ def _cast_to_dtype(value: VariableData, dtype: Optional[DType] = None) -> tf.Ten
         return tf.convert_to_tensor(value, dtype=dtype)
 
 
-def _to_constrained(value: VariableData, transform: Transform) -> tf.Tensor:
+def _to_constrained(value: TensorType, transform: Optional[Transform]) -> TensorType:
     if transform is not None:
         return transform.forward(value)
     return value
 
 
-def _to_unconstrained(value: VariableData, transform: Transform) -> tf.Tensor:
+def _to_unconstrained(value: TensorType, transform: Optional[Transform]) -> TensorType:
     if transform is not None:
         return transform.inverse(value)
     return value
